(*
 * Copyright (c) 2009, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import sepalg.
Require Import sepalg_generators.
Require Import sepalg_functors.
Require Import age_sepalg.
Require Import predicates_hered.

Require Import Wellfounded.
Require Import Recdef.

Delimit Scope pred with pred.
Local Open Scope pred.

Set Implicit Arguments.

Definition contractive `{ASA A} (f: pred A -> pred A) : Prop :=
  forall P Q,  |> (P <=> Q)  |-- f P <=> f Q.

Definition nonexpansive `{ASA A} (f: pred A -> pred A) : Prop :=
  forall P Q,  (P <=> Q)  |-- f P <=> f Q.

Definition HOcontractive `{ASA A} (X: Type) (f: (X -> pred A) -> (X -> pred A)) : Prop :=
  forall P Q,  (All_ x:X, |> (P x <=> Q x)) |-- (All_ x:X, f P x <=> f Q x).

Definition HOnonexpansive `{ASA A} (X: Type) (f: (X -> pred A) -> (X -> pred A)) : Prop :=
  forall P Q, (All_ x:X, P x <=> Q x)  |-- (All_ x:X, f P x <=> f Q x).

Module Type HO_REC.

  Parameter HORec : forall `{ASA A} X f (H:HOcontractive (X:=X) f), X -> pred A.
  Axiom HORec_fold_unfold : forall `{ASA A} X f (H:HOcontractive (X:=X) f),
    HORec H = f (HORec H).

  Parameter Rec : forall `{ASA A} f (H:contractive f), pred A.
  Axiom Rec_fold_unfold : forall `{ASA A} f (H:contractive f),
    Rec H = f (Rec H).

End HO_REC.

Module HoRec : HO_REC.

Section HORec.
  Variable A:Type.
  Variable sa: sepalg A.
  Variable sl : ASA sa.
  Variable X:Type.
  Variable f: (X-> pred A) -> (X -> pred A).

  Function HORec' (a:A) {wf (transp A age) a} : X -> pred A :=
    match age1 a with
    | Some a' => f (HORec' a')
    | None    => f (fun _ => FF)
    end.
    auto.
    apply (@asf_wf _ _ _ _ as_facts).
  Qed.

  Hypothesis Hcont : HOcontractive f.

  Lemma HORec'_unage : forall a1 a2 a1' a2' x a,
    fashionable a2 a2' ->
    necR a2' a1 ->
    fashionable a1 a1' ->
    necR a1' a ->
    (HORec' a1 x a <-> @HORec' a2 x a).
  Proof.
    intros a1.
    elim (wf_clos_trans _ _ (asf_wf as_facts) a1); clear a1; intros.
    rewrite (HORec'_equation x).
    case_eq (age1 x); intros.
    rewrite (HORec'_equation a2).
    case_eq (age1 a2); intros.

    destruct (@Hcont (HORec' a0) (HORec' a1)) with a x0 a.
    repeat (hnf; intros); simpl in *.
    
    assert (clos_trans A (transp _ age) a0 x).
    rewrite transpose_clos_trans.
    apply t_step; auto.
    spec H0 a0 H9 a1.
    cut (exists z1, fashionable a1 z1 /\ necR z1 a0).
    intros [z1 [? ?]].
    cut (exists w, fashionable w a' /\ necR a0 w).
    intros [w [? ?]].

    split; hnf; intros.
    cut (exists z2, fashionable a0 z2 /\ necR z2 a'0).
    intros [z2 [? ?]].
    spec H0 z2 z1 b a'1.
    destruct H0; auto.
    apply rt_trans with a'0; auto.
    destruct (valid_rel_commut_nec2 fashionR valid_rel_fashion) with a'0 w a0
      as [z ? ?]; auto.
    apply fashionable_trans with a'; auto.
    exists z; split; auto.
    cut (exists z2, fashionable a0 z2 /\ necR z2 a'0).
    intros [z2 [? ?]].
    spec H0 z2 z1 b a'1.
    destruct H0; auto.
    apply rt_trans with a'0; auto.
    destruct (valid_rel_commut_nec2 fashionR valid_rel_fashion) with a'0 w a0
      as [z ? ?]; auto.
    apply fashionable_trans with a'; auto.
    exists z; split; auto.

    destruct (valid_rel_commut_later1 fashionR valid_rel_fashion) with a' a1' x
      as [z ? ?]; auto.
    apply Rft_Rt_trans with a; auto.
    exists z; split; auto.
    apply age_later_nec with x; auto.

    apply nec_refl_or_later in H2.
    destruct H2.
    subst.
    exists a0.
    split.
    inv H1.
    rewrite H5 in H10; inv H10.
    rewrite H5 in H10; inv H10.
    rewrite H6 in H2; inv H2.
    auto.
    apply rt_refl.

    case_eq (age1 a2'); intros.
    exists a3; split.
    inv H1.
    rewrite H10 in H12; inv H12.
    rewrite H10 in H12; inv H12.
    rewrite H6 in H11; inv H11.
    auto.

    apply age_later_nec with a2'; auto.
    apply t_trans with x; auto.
    apply t_step; auto.
    elimtype False.
    clear -H10 H2.
    induction H2; auto.
    unfold age in H.
    rewrite H10 in H.
    discriminate.
    simpl.
    apply fashionable_refl.
    intuition.

    case_eq (age1 a2'); intros.
    inv H1.
    rewrite H7 in H9; inv H9.
    rewrite H6 in H8; inv H8.
    assert (laterR a2' a0).
    apply Rft_Rt_trans with x; auto.
    apply t_step; auto.
    elimtype False.
    clear -H8 H7.
    induction H8; auto.
    unfold age in *.
    rewrite H in H7; discriminate.
    rewrite (HORec'_equation a2).
    case_eq (age1 a2).
    2: intuition.
    intros.
    
    destruct (Hcont (fun _ => FF) (HORec' a0)) with a x0 a.
    hnf; intros; split; hnf; intros.
    elim H10.
    simpl in *.
    case_eq (age1 a1'); intros.
    inv H3.
    rewrite H11 in H13; inv H13.
    rewrite H5 in H12; inv H12.
    assert (laterR a1' a').
    apply Rft_Rt_trans with a; auto.
    elimtype False.
    clear -H12 H11.
    induction H12; auto.
    unfold age in *.
    rewrite H in H11; discriminate.
    simpl; apply fashionable_refl.
    intuition.
  Qed.

  Program Definition HORec_inner (x:X) : pred A := fun a => HORec' a x a.
  Next Obligation.
    hnf in Hcont.
    rewrite HORec'_equation in H0.
    hnf in H.
    rewrite H in H0.
    rewrite HORec'_equation.
    case_eq (age1 a'); intros.
    assert (f (HORec' a') x a').
    apply pred_hereditary with a; auto.
    generalize (Hcont (HORec' a') (HORec' a0)); intros.
    destruct H3 with a' x a'; auto.
    2: hnf; auto.
    
    clear H3.
    intro.
    rewrite later_age.
    hnf; simpl; intros.
    hnf in H3.
    rewrite H1 in H3; inv H3.
    split; intros.
    rewrite <- HORec'_unage in H5.
    apply H5.
    apply fashionable_refl.
    apply rt_step.
    hnf; auto.
    apply H4.
    auto.
    rewrite HORec'_unage in H5.
    apply H5.
    apply fashionable_refl.
    apply rt_step.
    auto.
    apply H4.
    auto.
    apply fashionable_refl.

    generalize (Hcont (HORec' a') (fun _ => FF)).
    intros.
    destruct H2 with a' x a'; auto.
    2: hnf; auto.
    hnf; simpl; intuition.
    clear -H1 H3.
    induction H3; auto.
    hnf in H; rewrite H1 in H; discriminate.
    apply fashionable_refl.
    apply H3; auto.
    apply pred_hereditary with a; auto.
  Qed.

  Lemma HORec_fold_unfold_inner :
    HORec_inner = f (HORec_inner).
  Proof.
    extensionality x.
    cut (forall a, HORec_inner x a <-> f (HORec_inner) x a).
    intros; apply equiv_eq; hnf; firstorder.

    intro a; simpl.
    rewrite (HORec'_equation).
    case_eq (age1 a); intros.
    destruct
      (@Hcont (HORec' a0)
              (HORec_inner) a) with x a.
    repeat (hnf; intros); split; hnf; simpl; intros.
    simpl in H1.
    destruct (valid_rel_commut_nec2 fashionR valid_rel_fashion) with a'0 a' a0
      as [z ? ?]; auto.
    apply age_later_nec with a; auto.
    rewrite -> (@HORec'_unage a'1 a0 a'1 z); auto.
    apply rt_trans with a'0; auto.
    apply fashionable_refl.
    destruct (valid_rel_commut_nec2 fashionR valid_rel_fashion) with a'0 a' a0
      as [z ? ?]; auto.
    apply age_later_nec with a; auto.
    rewrite <- (@HORec'_unage a'1 a0 a'1 z); auto.
    apply rt_trans with a'0; auto.
    apply fashionable_refl.
    simpl; apply fashionable_refl.
    intuition.

    destruct (@Hcont (fun _ => FF) (HORec_inner) a) with x a.
    repeat (hnf; intros); split; hnf; simpl; intros.
    elim H3.
    clear -H H0.
    induction H0; auto.
    hnf in H0.
    rewrite H in H0; discriminate.
    simpl; apply fashionable_refl.

    intuition.
  Qed.
End HORec.

Definition HORec `{sa:ASA A} : forall X f (H:HOcontractive f), X -> pred A
  := fun X f H x => @HORec_inner A _ sa X f H x.

Lemma HORec_fold_unfold `{ASA A} : forall X f (H:HOcontractive (X:=X) f),
  HORec H = f (HORec H).
Proof.
  intros.
  unfold HORec.
  apply HORec_fold_unfold_inner; auto.
Qed.


Section recursive.
  Variable A:Type.
  Variable sa:sepalg A.
  Variable asa:ASA sa.

  Variable f:pred A -> pred A.
  Variable Hc : contractive f.

  Lemma cont_HOcont : @HOcontractive A sa asa unit (fun x _ => f (x tt)).
  Proof.
    repeat intro.
    spec H tt.
    eapply Hc; eauto.
  Qed.
End recursive.


Definition Rec `{ASA A} f (H:contractive f) : pred A 
  := HORec (cont_HOcont H) tt.

Lemma Rec_fold_unfold : forall `{ASA A} f (H:contractive f),
  Rec H = f (Rec H).
Proof.
  intros.
  unfold Rec.
  pattern (HORec (cont_HOcont H0)) at 1.
  rewrite HORec_fold_unfold.
  auto.
Qed.

End HoRec.

Export HoRec.
