(*
 * Copyright (c) 2009, Andrew Appel, Robert Dockins and Aquinas Hobor.
 *
 *)

Require Import base.
Require Import sepalg.
Require Import sepalg_generators.
Require Import sepalg_functors.
Require Import age_sepalg.
Require Import predicates_hered.

Local Open Scope pred.

Lemma sub_equ `{ASA A} : forall G P Q,
  G |-- P >=> Q ->
  G |-- Q >=> P ->
  G |-- P <=> Q.
Proof.
  repeat intro.
  split.
  eapply H0; eauto.
  eapply H1; eauto.
Qed.

Lemma equ_sub `{ASA A} : forall G P Q,
  G |-- P <=> Q ->
  G |-- P >=> Q.
Proof.
  repeat intro.
  destruct (H0 a H1 a' H2); eauto.
Qed.

Lemma equ_sub2 `{ASA A} : forall G P Q,
  G |-- P <=> Q ->
  G |-- Q >=> P.
Proof.
  repeat intro.
  destruct (H0 a H1 a' H2); eauto.
Qed.

Lemma sub_refl `{ASA A} : forall G P,
  G |-- P >=> P.
Proof.
  repeat intro; auto.
Qed.

Lemma sub_trans `{ASA A} : forall G P Q R,
  G |-- P >=> Q ->
  G |-- Q >=> R ->
  G |-- P >=> R.
Proof.
  repeat intro.
  eapply H1; eauto.
  eapply H0; eauto.
Qed.

Lemma sub_top `{ASA A} : forall G P,
  G |-- P >=> TT.
Proof.
  repeat intro; simpl; auto.
Qed.

Lemma sub_bot `{ASA A} : forall G P,
  G |-- FF >=> P.
Proof.
  repeat intro; simpl in *; intuition.
Qed.

Lemma sub_conj `{ASA A} : forall G P P' Q Q',
  G |-- P >=> P' ->
  G |-- Q >=> Q' ->
  G |-- P && Q >=> (P' && Q').
Proof.
  repeat intro.
  destruct H5; split.
  eapply H0; eauto.
  eapply H1; eauto.
Qed.

Lemma sub_impl `{ASA A} : forall G P P' Q Q',
  G |-- P' >=> P ->
  G |-- Q >=> Q' ->
  G |-- (P --> Q) >=> (P' --> Q').
Proof.
  repeat intro.
  eapply H1; eauto.
  apply H5; auto.
  eapply H0; eauto.
Qed.

Lemma sub_disj `{ASA A} : forall G P P' Q Q',
  G |-- P >=> P' ->
  G |-- Q >=> Q' ->
  G |-- (P || Q) >=> (P' || Q').
Proof.
  repeat intro.
  destruct H5; [ left | right ].
  eapply H0; eauto.
  eapply H1; eauto.
Qed.

Lemma sub_forall `{ASA A} : forall G B (X Y:B -> pred A),
  (forall x:B, G |-- X x >=> Y x) ->
  G |-- universal X >=> universal Y.
Proof.
  repeat intro.
  eapply H0; eauto.
Qed.

Lemma sub_exists `{ASA A} : forall G B (X Y:B -> pred A),
  (forall x:B, G |-- X x >=> Y x) ->
  G |-- existential X >=> existential Y.
Proof.
  repeat intro.
  destruct H4; exists x.
  eapply H0; eauto.
Qed.

Lemma sub_forall_spec `{ASA A} : forall G B (X:B -> pred A) x,
  G |-- universal X >=> X x.
Proof.
  repeat intro; eauto.
Qed.

Lemma sub_exists_spec `{ASA A} : forall G B(X:B -> pred A) x,
  G |-- X x >=> existential X.
Proof.
  repeat intro.
  exists x; auto.
Qed.

Lemma sub_later `{ASA A} : forall G P Q,
  G |-- |>(P >=> Q) <-> G |-- |>P >=> |>Q.
Proof.
  intros.
  rewrite <- later_commute.
  rewrite later_impl.
  tauto.
Qed.

Lemma sub_box `{ASA A} : forall G (M:modality) P Q,
  inclusion A M fashionable ->
  G |-- P >=> Q ->
  G |-- box M P >=> box M Q.
Proof.
  repeat intro.
  generalize (H0 _ _ H6); intros.
  rewrite <- (box_refl_trans fashionM) in H1.
  apply H1 in H2.
  spec H2 a' H3.
  eapply pred_nec_hereditary in H2; eauto.
  spec H2 a'1 H7.
  apply H2; auto.
  simpl; apply fashionable_refl.
  simpl; apply fashionable_trans.
Qed.

Lemma sub_diamond `{ASA A} : forall G (M:modality) P Q,
  inclusion A M fashionable ->
  G |-- P >=> Q ->
  G |-- diamond M P >=> diamond M Q.
Proof.
  repeat intro.
  destruct H5 as [w [? ?]].
  exists w; split; auto.
  rewrite <- (box_refl_trans fashionM) in H1.
  apply H1 in H2.
  spec H2 a' H3.
  eapply pred_nec_hereditary in H2; eauto.
  assert (fashionable a'0 w).
  apply fasionable_sym.
  apply H0; auto.
  eapply H2; eauto.
  simpl; apply fashionable_refl.
  simpl; apply fashionable_trans.
Qed.

Lemma sub_con `{ASA A} : forall G P P' Q Q',
  G |-- P >=> P' ->
  G |-- Q >=> Q' ->
  G |-- P * Q >=> P' * Q'.
Proof.
  repeat intro.
  destruct H5 as [w1 [w2 [? [? ?]]]].
  exists w1; exists w2; split; auto.
  split.
  spec H0 a H2.
  rewrite <- (box_refl_trans fashionM) in H0.
  spec H0 a' H3.
  eapply pred_nec_hereditary in H0; eauto.
  eapply H0; eauto.
  simpl; apply comparable_fashionable.
  apply comparable_sym.
  eapply join_sub_comparable; eauto.
  simpl; apply fashionable_refl.
  simpl; apply fashionable_trans.
  spec H1 a H2.
  rewrite <- (box_refl_trans fashionM) in H1.
  spec H1 a' H3.
  eapply pred_nec_hereditary in H1; eauto.
  eapply H1; eauto.
  simpl; apply comparable_fashionable.
  apply comparable_sym.
  eapply join_sub_comparable; eauto.
  apply join_com; eauto.
  simpl; apply fashionable_refl.
  simpl; apply fashionable_trans.
Qed.

Lemma sub_wand `{ASA A} : forall G P P' Q Q',
  G |-- P' >=> P ->
  G |-- Q >=> Q' ->
  G |-- (P -* Q) >=> (P' -* Q').
Proof.
  repeat intro.
  spec H1 a H2.
  rewrite <- (box_refl_trans fashionM) in H1.
  spec H1 a' H3.
  eapply pred_nec_hereditary in H1; eauto.
  eapply pred_nec_hereditary in H1; eauto.
  eapply H1; eauto.
  simpl; apply comparable_fashionable.
  eapply join_sub_comparable; eauto.
  eapply H5; eauto.
  spec H0 a H2.
  rewrite <- (box_refl_trans fashionM) in H0.
  spec H0 a' H3.
  eapply pred_nec_hereditary in H0; eauto.
  eapply pred_nec_hereditary in H0; eauto.
  eapply H0; eauto.
  simpl; apply comparable_fashionable.
  eapply join_comparable; eauto.
  simpl; apply fashionable_refl.
  simpl; apply fashionable_trans.
  simpl; apply fashionable_refl.
  simpl; apply fashionable_trans.
Qed.
  
